@@ -98,10 +98,14 @@ func (s Sum) String() string {
 }
 
 var (
-	once sync.Once
-	seed uint64
+	seedOnce sync.Once
+	seed     uint64
 )
 
+func initSeed() {
+	seed = uint64(time.Now().UnixNano())
+}
+
 func (h *hasher) sum() (s Sum) {
 	h.bw.Flush()
 	// Sum into scratch & copy out, as hash.Hash is an interface
@@ -121,9 +125,7 @@ func Hash(v any) (s Sum) {
 	h := hasherPool.Get().(*hasher)
 	defer hasherPool.Put(h)
 	h.reset()
-	once.Do(func() {
-		seed = uint64(time.Now().UnixNano())
-	})
+	seedOnce.Do(initSeed)
 	h.hashUint64(seed)
 	h.hashValue(reflect.ValueOf(v))
 	return h.sum()
@@ -298,9 +300,9 @@ func (h *hasher) hashValue(v reflect.Value) {
 }
 
 type mapHasher struct {
-	h    hasher
-	val  valueCache      // re-usable values for map iteration
-	iter reflect.MapIter // re-usable map iterator
+	h               hasher
+	valKey, valElem valueCache      // re-usable values for map iteration
+	iter            reflect.MapIter // re-usable map iterator
 }
 
 var mapHasherPool = &sync.Pool{
@@ -334,15 +336,19 @@ func (h *hasher) hashMap(v reflect.Value) {
 	defer iter.Reset(reflect.Value{}) // avoid pinning v from mh.iter when we return
 
 	var sum Sum
-	k := mh.val.get(v.Type().Key())
-	e := mh.val.get(v.Type().Elem())
+	if v.IsNil() {
+		sum.sum[0] = 1 // something non-zero
+	}
+
+	k := mh.valKey.get(v.Type().Key())
+	e := mh.valElem.get(v.Type().Elem())
 	mh.h.visitStack = h.visitStack // always use the parent's visit stack to avoid cycles
 	for iter.Next() {
 		k.SetIterKey(iter)
 		e.SetIterValue(iter)
 		mh.h.reset()
 		mh.h.hashValue(k)
-		mh.h.hashValue(v)
+		mh.h.hashValue(e)
 		sum.xor(mh.h.sum())
 	}
 	h.bw.Write(append(h.scratch[:0], sum.sum[:]...)) // append into scratch to avoid heap allocation
